/*
 *  This code is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  This code is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#include "DVDPreview.h"
#include <stdio.h>

// All these are defined in the DVD2BMP code...
// This is the main preview window destroy function. Pass the handle
extern BOOL DestroyMain(HWND TheHWnd);
// This is the main preview initialization function. Pass the handle
// of the parent window, and the desired initial X and Y positions of the preview window in the parent window.
extern HWND InitMain(HWND TheHWnd, int PosX, int PosY, UINT ShowWindow);
// You can call this directly, instead of sending messages. This is how you control Preview.
extern LRESULT CALLBACK TheWndProc(HWND, UINT, WPARAM, LPARAM);
// This loads a file. LoadMany should be 0. Returns 0 if all goes well.
extern int PreviewLoadFile(const char* FileName, int LoadMany);
extern void PreviewCloseAll();
extern void PreviewSetVideoPosition(int PositionX, int PositionY);
// Use this to force the x-axis and y-axis scaling factor irrespective of any DAR info, or the
// size of the coded video.
extern void SetVideoScalingFactor(float XScale, float YScale);
// Use this to simply set an x-axis scaling factor, and using the DAR and the video coding size
// to determine the y-axis scaling factor. If DAR == -1, the DAR in the video stream is used.
// If DAR is 0,1,2,3 the corresponding DAR is used:
// 0 and 1 are "1:1", 2 is "4:3", 3 is "16:9", 4 is "2.21:1"
extern void SetVideoScalingFactor(float XScale, int DAR);
// Because Preview is multithreaded, you need to wait for it to finish when you send commands or
// call TheWndProc() so your subsequent messages are not ignored. This function will do the trick.
extern void PreviewWaitForThread();
// The window proc function for the video window. I'm using it to send messages directly, bypassing windows.
extern LRESULT CALLBACK DVD2AVIWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
// This function is for opening a vob file, without displaying anything and saving a bitmap.
extern PreviewSaveToBMPFile(char* VideoFileName, int Sector, float Xfactor, float Yfactor, char* BMPFileName);
// This refreshes the window, but if Display_Flag isn't true, nothing will be shown...
extern void RefreshWindow(int);

extern "C" void ResizeWindow(int width, int height);
// A quick release function.
extern void ReleaseAll();
// This updates all the info in the info panel.
extern "C" void UpdateInfo();
// This erases the video frame.
extern "C" void EraseFrame(COLORREF Color);

extern "C" HWND hWnd;					// Handle of video windows.
extern UINT WindowStyle;				// The style of the video window.
extern "C" float XDownSampFactor;		// X Image downsampling factor for bitmap file saving.
extern "C" float YDownSampFactor;		// Y Image downsampling factor for bitmap file saving.
extern "C" int Display_Flag;			// When true, the video is shown. 
//extern "C" int Coded_Picture_Width, Coded_Picture_Height;	// Size of DVD picture.
extern "C" int Coded_Picture_Width, Vert_Display_Size;	// Size of DVD picture.
extern int	DisplayVideoOnly;			// If true, there's no slider, and no buttons!
extern HANDLE hThread;					// The DVD2AVI thread.
extern int	LastFileSector;				// Last sector of VOB file.
extern "C" int CumulStartSectors[MAX_FILE_NUMBER]; // This stores the cumulated start sectors for each VOB file.
extern "C" int Stop_LBA ;				// A sector where playback will stop if not -1.
extern "C" BOOL Fps_Playback;			// When true, playback will fit fps rate. 
extern "C" int Fault_Flag;				// DVD2AVI's error flag. Should be 0 if all goes well.
extern "C" int NeverUseOverlay;			// No direct draw overlay calls if that's not 0.
extern "C" float MaxVideoScale;			// Maximum scaling factor for the video.
extern "C" HDC hDC;
extern "C" int WSDisplayMode;	// In bitmap mode only, controls the 16:9 display mode: 0 for no scaling
// 1 for letterbox and 2 for pan&scan.


extern int InfoXPos;	// Position of info panel.
extern int InfoYPos;	// Position of info panel.
extern int FloatingInfo;// Is the info window "floating"? (moves with the video window).
extern char BmpFileName[256];			// File name for bitmap file saving.
extern int	SaveAsPPM;					// If 1, bitmap saving saves a ppm file instead of a bmp.

extern "C" char Num_Buttons;	// Number of buttons defined in current nav-pack.
extern "C" int VOB_ID, CELL_ID;	// VOB ID and CELL ID in current nav-pack..
extern "C" int DrawButtons ;		// If 1, button rectangles are drawn.
extern "C" COLORREF ButtonColor;	// Color used to draw button rectangles.

extern "C" void (*PostDisplay)(HDC,void*);
extern "C" void* PostDisplayParam;
extern "C" void (*BitmapMixFunction)(unsigned char*, int, int, void*, int);
extern "C" void* BitmapMixParam;

// This is defined in global.h, but I can't include global.h because there's a lot of variables declared
// there! This is ugly, I have to replicate the definition of this structure...
typedef struct {
	int			type;
	int			file;
	__int64		lba;
	bool		pf;
	bool		trf;
}	D2VData;
D2VData;
extern "C" D2VData d2v_current;		// I'm using it to get the current sector (lba).

HRESULT (_stdcall *pDwmIsCompositionEnabled)(BOOL *pfEnabled);

//int __stdcall AfxMessageBox(LPCTSTR lpszText, UINT nType = MB_OK,
//				UINT nIDHelp = 0);

BOOL NoOverlayBecauseOfVista = FALSE;

CDVDPreview::CDVDPreview()
{
	// In vista, if you don't disable the overlay while the desktop window manager is allowing composition, things
	// go bad. Load library, get the function call address, then make the call. This isn't very robust though... :(
	HINSTANCE DWMAPI = LoadLibrary("dwmapi.dll");
//	char Temp[1024];
//	sprintf(Temp,"Found dwmapi.dll = %d",DWMAPI);
//	AfxMessageBox(Temp,MB_OK,0);
	if(DWMAPI != 0)
	{
		// _cdecl does NOT work in VISTA (crashes).
//		pDwmIsCompositionEnabled = (HRESULT (_cdecl*)(BOOL*)) GetProcAddress(DWMAPI,"DwmIsCompositionEnabled");
		pDwmIsCompositionEnabled = (HRESULT (_stdcall*)(BOOL*)) GetProcAddress(DWMAPI,"DwmIsCompositionEnabled");
//		sprintf(Temp,"Found DwmIsCompositionEnabled = %d",pDwmIsCompositionEnabled);
//		AfxMessageBox(Temp,MB_OK,0);
		BOOL Ret = -1;
		if(pDwmIsCompositionEnabled)
		{
			HRESULT Fuck = pDwmIsCompositionEnabled(&NoOverlayBecauseOfVista);
//			sprintf(Temp,"DwmIsCompositionEnabled returned = %d NoOverlayBecauseOfVista = %d",Fuck,NoOverlayBecauseOfVista);
//			AfxMessageBox(Temp,MB_OK,0);
		}
		FreeLibrary(DWMAPI);
	}
}


CDVDPreview::~CDVDPreview()
{
}

void CDVDPreview::PlaybackActualRate(BOOL Fps)
{
	Fps_Playback=Fps;
}

BOOL CDVDPreview::DestroyPreview(HWND ParentWindow)
{
	// Call the main windoe destroy function.
	return DestroyMain(ParentWindow);
}

HWND CDVDPreview::InitPreview(HWND ParentWindow, float VideoScaleX, float VideoScaleY, int PosX, int PosY,
								UINT ShowWindow, BOOL VideoOnly, UINT AWindowStyle, BOOL NoOverlay)
{
	DisplayVideoOnly = VideoOnly;
	NeverUseOverlay = (int) (NoOverlay || NoOverlayBecauseOfVista);
	if(VideoScaleY == -1)
		// Autoscaling mode. Scaling according to DAR in video.
		SetVideoScalingFactor(VideoScaleX, -1);
	else
		// Strict scaling mode, DAR ignored.
		SetVideoScalingFactor(VideoScaleX, VideoScaleY);
	
	WindowStyle = AWindowStyle;
	// Call the main initialization function.
	return InitMain(ParentWindow, PosX, PosY, ShowWindow);
}

int CDVDPreview::LoadFile(const char* FileName, bool LoadAll, bool Refresh)
{
	int ret = PreviewLoadFile(FileName, LoadAll);
	if(ret != 0)
		return ret;
	WaitForThread();
	// Now refresh the video (or you'll still see the old frame).
	RefreshVideo(Refresh);
	return ret;
}

void CDVDPreview::CloseFile()
{
	PreviewCloseAll();
}

int CDVDPreview::CheckErrorFlag()
{
	return Fault_Flag;
}


void CDVDPreview::RefreshVideo(bool Refresh)
{
	// I have to set this flag to make sure the video will indeed be shown!
	// Sometimes, it's useful to refresh the video but still not show anything, the optional
	// argument to this function allows that.
	Display_Flag = Refresh;
	RefreshWindow(true);
}

void CDVDPreview::RePaintVideo()
{
	RefreshWindow(0);
}


void CDVDPreview::SetMaxVideoScaling(float MaxScaling)
{
	MaxVideoScale = MaxScaling;
}


void CDVDPreview::SetVideoScaling(float VideoScaleX, float VideoScaleY, bool Refresh)
{
	if(VideoScaleY == -1)
	{
		// If VideoScaleY we're using the aspect ratio stored in the video file to calculate
		// the y-axis scaling factor.
		SetVideoScalingFactor(VideoScaleX, -1);
	}
	else
	{
		SetVideoScalingFactor(VideoScaleX, VideoScaleY);
	}
	if(!Refresh)
		return;
	// If no file is open, refreshing the window won't resize the view.
	// This forces it...
	Sleep(30);
	ResizeWindow(Coded_Picture_Width,Vert_Display_Size);
	// This will refresh the window...
	RefreshWindow(1);
}

void CDVDPreview::SetVideoScaling(float VideoScaleX, int DAR)
{
	SetVideoScalingFactor(VideoScaleX, DAR);
	// If no file is open, refreshing the window won't resize the view.
	// This forces it...
	Sleep(30);
	ResizeWindow(Coded_Picture_Width,Vert_Display_Size);
	// This will refresh the window...
	RefreshWindow(1);
}

void CDVDPreview::SetWSDisplayMode(int letterbox)
{
	WSDisplayMode = letterbox;
}


void CDVDPreview::ShowSliderAndButtons(bool Show)
{
	DisplayVideoOnly = !Show;
	ResizeWindow(Coded_Picture_Width,Vert_Display_Size);
	// This will refresh the window...
	RefreshWindow(1);
}

LRESULT CDVDPreview::HandleMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
	// Simply calling the video window's proc function on the given message.
	return DVD2AVIWndProc(hWnd, message, wParam, lParam);
}

void CDVDPreview::WaitForThread()
{
	// Because the original DVD2AVI code is multithreaded, most of the calls we make to it will be
	// executed in another thread and return immediately. This will wait for that other thread to return.
	PreviewWaitForThread();
}

void CDVDPreview::SaveBMPFile(char* FileName, float XDownSamp, float YDownSamp, int SavePPM)
{
	XDownSampFactor = XDownSamp;
	YDownSampFactor = YDownSamp;
	if(FileName != NULL)
		strcpy(BmpFileName,FileName);
	else
		strcpy(BmpFileName,"");

	SaveAsPPM = SavePPM;
	// "Send" a IDM_BMP message. This will cause a "Save" dialog to pop up and the BMP file will be written.
	HandleMessage(WM_COMMAND, IDM_BMP,0);
}

int	CDVDPreview::SaveToBMPFile(char* VideoFileName, int Sector, float Xfactor, float Yfactor, 
		char* BMPFileName)
{
	// This call PreviewSaveToBMPFile() defined in gui.cpp, which handles everything. There's no "Save"
	// dialog opened. BMPFileName is used instead. 
	return PreviewSaveToBMPFile(VideoFileName, Sector, Xfactor, Yfactor, BMPFileName);
}


void CDVDPreview::Play()
{
	HandleMessage(WM_COMMAND,IDM_PLAY,0);
}

void CDVDPreview::Stop()
{
	HandleMessage(WM_COMMAND,IDM_STOP,0);
}

long CDVDPreview::GoToSector(long SectorNum, int Display)
{
	long CurrentSector;
	SectorNum = max(0,SectorNum);
	SectorNum = min(LastFileSector,SectorNum);
	CurrentSector = SectorNum;
	Display_Flag = Display;
	HandleMessage(WM_USER,ID_USERPOSITION,SectorNum);
	return CurrentSector;
}

bool CDVDPreview::IsVideoRunning()
{
	// This returns immediately, indicating wether hThread is running (video is playing, typically).
	return WaitForSingleObject(hThread, 0)!=WAIT_OBJECT_0;
}

long CDVDPreview::GetCurrentSector()
{
	// d2v_current.lba will be 0 when we're at the start of a VOB file. CurrentSector is the cumulated
	// sector so we have to add the cumulated sector at the start of the current VOB file.
	long CurrentSector;
	CurrentSector = (long) d2v_current.lba + CumulStartSectors[d2v_current.file];
	return 	CurrentSector;
}

long CDVDPreview::GetLastSector()
{
	// Last sector in last VOB file.
	return LastFileSector;
}

void CDVDPreview::GoToNextVideoFrame(bool Display)
{
	// I'm passing the display argument in the lparam.
	HandleMessage(WM_KEYDOWN,VK_RIGHT,(UINT)(!Display));
}

void CDVDPreview::GoToPreviousVideoFrame(bool Display)
{
	// I'm passing the display argument in the lparam.
	HandleMessage(WM_KEYDOWN,VK_LEFT,(UINT)(!Display));
}

void CDVDPreview::SetStopSector(int Sector)
{
	Stop_LBA = Sector;
}

void CDVDPreview::QuickRelease()
{
	Stop();
	WaitForThread();
	ReleaseAll();
}

void CDVDPreview::SetVideoPosition(int PositionX, int PositionY, bool Refresh)
{
	PreviewSetVideoPosition(PositionX, PositionY);
	// This will refresh the window...
	if(Refresh)
		RefreshWindow(1);
}

void CDVDPreview::ShowInfo(bool Show)
{
	// Send message, then force an update of the info...
	HandleMessage(WM_COMMAND,IDM_INFO,Show);
	if(Show)
		UpdateInfo();
}

void CDVDPreview::SetInfoPosition(int XPos, int YPos)
{
	InfoXPos = XPos;
	InfoYPos = YPos;
	FloatingInfo = 0;
}

void CDVDPreview::GetInfoPosition(int &XPos, int &YPos)
{
	XPos = InfoXPos;
	YPos = InfoYPos;
}

void CDVDPreview::GetVobCellID(int &VobID, int &CellID)
{
	VobID = VOB_ID;
	CellID = CELL_ID;
}

int CDVDPreview::GetNumButtons()
{
	return Num_Buttons;	
}

void CDVDPreview::SetDisplayMode(int Display)
{
	Display_Flag = Display;
}

int CDVDPreview::GetCodedPictureWidth()
{
	return Coded_Picture_Width;
}	

int CDVDPreview::GetCodedPictureHeight()
{
	return Vert_Display_Size;
}

void CDVDPreview::SetDrawButtons(bool DrawThem, COLORREF Color)
{
	DrawButtons = DrawThem ? 1 : 0;
	ButtonColor = Color;
}	

bool CDVDPreview::GetDrawButtons()
{
	return DrawButtons != 0;
}

void CDVDPreview::EraseVideoFrame(COLORREF Color)
{
	EraseFrame(Color);
}

HDC CDVDPreview::GetDC()
{
	return hDC;
}

void CDVDPreview::SetPostDisplayFunc(void (*PostDFun)(HDC,void*),void* Param)
{
	PostDisplayParam = Param;
	PostDisplay = PostDFun;
}

void CDVDPreview::SetBitmapMixFunction(void (*BitmapMFunc)(unsigned char*, int, int, void*, int),void* Param)
{
	BitmapMixParam = Param;
	BitmapMixFunction = BitmapMFunc;
}


// This function tests whether a file contains a valid, un-encrypted VOB, by looking for the
// 1BA value in the first 4 bytes of the first NUM_TEST_ENCODED_PACKS packs...
#define NUM_TEST_ENCODED_PACKS 100
int CDVDPreview::CheckFileEncoded(char* FileName)
{
	FILE* fd = fopen(FileName,"rb");
	if(fd == 0)
		return -1;
	
	DWORD LBA;
	int Encoded = 1;
	int i;
	for(i=0;i<NUM_TEST_ENCODED_PACKS;i++)
	{
		if(fread(&LBA,sizeof(DWORD),1,fd) != 1)
			break;
		// If it's 0, it might be one of these weird ARCCOS packs...
//		if(LBA != 0 && LBA != 0xba010000)
		if(LBA == 0xba010000)
			Encoded = 0;
		if(fseek(fd,i*1024*2,SEEK_SET) != 0)
			Encoded = 1;
	}
	fclose(fd);
	return Encoded;
}