/* 
 *	Copyright (C) Chia-chen Kuo - April 2001
 *
 *  This file is part of DVD2AVIMFC, a free MPEG-2 decoder
 *	
 *  DVD2AVIMFC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DVD2AVIMFC is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include "global.h"
#include "filter.h"

#define MAX_AVI_SIZE	2073600000

__forceinline static void Store_RGB24(unsigned char *src[], DWORD frame);
__forceinline static void Store_YUY2(unsigned char *src[], DWORD frame);


static void FlushRGB24(void);
static void FlushYUY2(void);
static void AVIKill(void);
static void SaveBMPToFile();

static PAVIFILE pfile;
static PAVISTREAM ps, psCompressed;
static AVICOMPRESSOPTIONS opts;
static AVISTREAMINFO strinfo;
static ICCOMPRESSFRAMES iccf;
static COMPVARS compvars;
static BITMAPINFOHEADER birgb, biyuv;

static bool AVI_Init;
static char VideoOut[_MAX_PATH];
static unsigned char *y444;
static int avi_size, avi_count, frame_count;
static int playback, Old_Playback;
static long frame_size;
static bool TFF, RFF, TFB, BFB, frame_type;

static bool Resize_Flag;
static int Resize_Width, Resize_Height;
float XDownSampFactor = 1; // X Image downsampling factor for bitmap file saving.
float YDownSampFactor = 1; // Y Image downsampling factor for bitmap file saving.
float YUVScaleX = 1;	// X scaling factor for YUV display. YUVScaleX > 1 means enlarging.
float YUVScaleY = 1;	// Y scaling factor for YUV display. 
int   AutoScale = 1;	// If 1, YUVScaleY is calculated from YUVScaleX and the aspect ratio.
int   ForcedDAR = -1;	// If different than -1, this DAR is used instead of the one in the video.
						// Valid values are starting from 0: "", "1:1", "4:3", "16:9", "2.21:1"
int   NeedsScaling = 1;	// Indicates whether the scaled bitmap is in sync with the original non-scaled one.

int   NeverUseOverlay = 1;		// No direct draw overlay calls if that's not 0.
COLORREF ButtonColor = RGB(255, 0, 0);	// Color used to draw button rectangles.

// These are here to provide a way to call a function right after a frame is displayed. If PostDisplay
// isn't NULL, the function is called, and the current HDC and PostDisplayParam are passed as arguments.
// This can be used, for example, to display subtitles on top of the video...
// BitmapFunction is called before the final image is rescaled, so you can mix in something if you want to.
void (*BitmapMixFunction)(unsigned char*, int, int, void*, int) = 0;
void (*PostDisplay)(HDC,void*) = 0;
void* PostDisplayParam = 0;
void* BitmapMixParam = 0;

static char *FrameType[] = {
	"Interlaced", "Progressive"
};


void Write_Frame(unsigned char *src[], D2VData d2v, DWORD frame)
{
	int repeat;
	static unsigned int timeOffset = 0;
	static float DisplayTime=0;
	static int elapsed=0;
	int sleep_time;

	if (Fault_Flag)
	{
		if (Fault_Flag < CRITICAL_ERROR_LEVEL)
		{
			SetDlgItemText(hDlg, IDC_INFO, "V.E.!");
			Fault_Flag = 0;		// fault tolerance
		}
		else
		{
			if (AVI_Flag)
				AVIKill();
			ThreadKill();
		}
	}

	frame_type = d2v.pf;
	TFF = d2v.trf>>1;
	RFF = d2v.trf & 0x01;

	if (!frame)
	{
		int i;

		AVI_Init = true; TFB = BFB = false; Old_Playback = 0;
		frame_size = 0; avi_size = 0; avi_count = 1; frame_count = 0; playback = 0;

		Resize_Flag = false;
		Clip_Width = Resize_Width = Coded_Picture_Width;
		Clip_Height = Resize_Height = Coded_Picture_Height;
		CLIP_AREA = HALF_CLIP_AREA = CLIP_STEP = CLIP_HALF_STEP = 0;

		if (Luminance_Flag)
			InitializeFilter();

		if (ClipResize_Flag)
		{
			if (Clip_Top || Clip_Bottom || Clip_Left || Clip_Right)
			{
				Clip_Width -= Clip_Left+Clip_Right;
				Clip_Height -= Clip_Top+Clip_Bottom;
				Resize_Width = Clip_Width;
				Resize_Height = Clip_Height;

				CLIP_AREA = Coded_Picture_Width * Clip_Top;
				HALF_CLIP_AREA = (Coded_Picture_Width>>1) * Clip_Top;

				CLIP_STEP = Coded_Picture_Width * Clip_Top + Clip_Left;
				CLIP_HALF_STEP = (Coded_Picture_Width>>1) * Clip_Top + (Clip_Left>>1);
			}

			if (Squeeze_Width || Squeeze_Height)
			{
				Resize_Flag = true;
				Resize_Width -= Squeeze_Width;
				Resize_Height -= Squeeze_Height;
			}
		}

		NINE_CLIP_WIDTH = Clip_Width * 9;
		QUAD_CLIP_WIDTH = Clip_Width<<2;
		DOUBLE_CLIP_WIDTH = Clip_Width<<1;
		HALF_CLIP_WIDTH = Clip_Width>>1;

		LUM_AREA = Coded_Picture_Width * Clip_Height;
		DOUBLE_WIDTH = Coded_Picture_Width<<1;
		HALF_WIDTH = Coded_Picture_Width>>1;
		HALF_WIDTH_D8 = (Coded_Picture_Width>>1) - 8;
		PROGRESSIVE_HEIGHT = (Coded_Picture_Height>>1) - 2;
		INTERLACED_HEIGHT = (Coded_Picture_Height>>2) - 2;
		RGB_DOWN1 = Clip_Width * (Clip_Height - 1) * 3;
		RGB_DOWN2 = Clip_Width * (Clip_Height - 2) * 3;

#if 0
		// Overriding that stuff too. JEANL I don't want stuff on the window title!
		sprintf(szBuffer, "DVD2AVIMFC - [%d / %d] ", File_Flag+1, File_Limit);
		ext = strrchr(Infilename[File_Flag], '\\');
		strncat(szBuffer, ext+1, strlen(Infilename[0])-(int)(ext-Infilename[0]));

		sprintf(szTemp, " %d x %d", Resize_Width, Resize_Height);
		strcat(szBuffer, szTemp);

		if (Resize_Width%32 == 0)
			strcat(szBuffer, " [32X]");
		else if (Resize_Width%16 == 0)
			strcat(szBuffer, " [16X]");
		else if (Resize_Width%8 == 0)
			strcat(szBuffer, " [8X]");
		else
			strcat(szBuffer, " [ ]");

		if (Resize_Height%32 == 0)
			strcat(szBuffer, "[32X]");
		else if (Resize_Height%16 == 0)
			strcat(szBuffer, "[16X]");
		else if (Resize_Height%8 == 0)
			strcat(szBuffer, "[8X]");
		else
			strcat(szBuffer, "[ ]");
// jsoto_start
		SetWindowText(hWnd, szBuffer);
// jsoto_end
#endif
		// I've moved that in the store() function (JEAN), so resizing can occur even if we're only refreshing
		// an already decoded picture.
//		ResizeWindow(Clip_Width, Vert_Display_Size);

		ZeroMemory(&birgb, sizeof(BITMAPINFOHEADER));
		birgb.biSize = sizeof(BITMAPINFOHEADER);
		birgb.biWidth = Clip_Width;
		birgb.biHeight = Clip_Height;
		birgb.biPlanes = 1;
		birgb.biBitCount = 24;
		birgb.biCompression = BI_RGB;
		birgb.biSizeImage = Clip_Width * Clip_Height * 3;

		ZeroMemory(&biyuv, sizeof(BITMAPINFOHEADER));
		biyuv = birgb;
		biyuv.biBitCount = 16;
		biyuv.biCompression = mmioFOURCC('Y','U','Y','2');
		biyuv.biSizeImage = Clip_Width * Clip_Height * 2;

		if (process.locate==LOCATE_RIP && FO_Flag==FO_SWAP)
		{
			if (TFF)
				BFB = true;
			else
				TFB = true;

			// mask
			ZeroMemory(rgb24, Coded_Picture_Width * Clip_Height * 3);
			for (i=0; i<Coded_Picture_Width * Coded_Picture_Height * 2; i+=2)
			{
				yuy2[i] = 0;
				yuy2[i+1] = 128;
			}
		}

		if (FO_Flag!=FO_FILM)
		{
			if (TFF && FO_Flag!=FO_SWAP  ||  !TFF && FO_Flag==FO_SWAP)
				SetDlgItemText(hDlg, IDC_INFO, "T");
			else
				SetDlgItemText(hDlg, IDC_INFO, "B");
		}
	}

	repeat = DetectVideoType(frame, d2v.trf);
	// We decoded a new picture, it will need to be scaled.
	NeedsScaling = 1;

	if (FO_Flag!=FO_FILM || repeat)
	{
		if (Store_Flag==STORE_YUY2)
			Store_YUY2(src, frame); 
		else
			Store_RGB24(src, frame);
	}

	if (FO_Flag==FO_FILM && repeat==2)
	{
		if (Store_Flag==STORE_YUY2)
			Store_YUY2(src, frame);
		else
			Store_RGB24(src, frame);
	}
//jsoto_start: Playback delay to fit the actual fps
	if (frame &&  Fps_Playback && process.locate==LOCATE_RIP)
	{
		elapsed = timeGetTime() - timeOffset;
		sleep_time = (int) (DisplayTime - elapsed +1);
		// Reset every 64 frames, but...
		// if sleep_time <0 means we just started to play (elapsed too large...)
		if ((frame & 63)==0 || sleep_time < 0 )
		{
			timeOffset=timeGetTime();
			DisplayTime=0;
		}
		// Just to guarantee a minimum/maximum
		// sleep_time (5msec=200 fps; 60msec = 16fps)
		if (sleep_time >60) sleep_time=60;
		if (sleep_time <5 ) sleep_time=5;
		Sleep(sleep_time);
		DisplayTime += (float) 1000.0 / frame_rate;
	}
//jsoto_end

	if (Info_Flag && process.locate==LOCATE_RIP)
	{
		sprintf(szBuffer, "%s", FrameType[frame_type]);
		SetDlgItemText(hDlg, IDC_FRAME_TYPE, szBuffer);

		sprintf(szBuffer, "%d", frame+1);
		SetDlgItemText(hDlg, IDC_CODED_NUMBER, szBuffer);

		sprintf(szBuffer, "%d", playback);
		SetDlgItemText(hDlg, IDC_PLAYBACK_NUMBER, szBuffer);

		if (AVI_Flag)
		{
			sprintf(szBuffer, "%d", avi_count-1);
			SetDlgItemText(hDlg, IDC_FILE, szBuffer);
	
			sprintf(szBuffer, "%d MB", avi_size>>20);
			SetDlgItemText(hDlg, IDC_FILE_SIZE, szBuffer);
		}

		if ((frame & 63) == 63)
		{
			process.ed = timeGetTime();

			sprintf(szBuffer, "%.1f", 1000.0*(playback-Old_Playback)/(process.ed-process.mi+1));
			SetDlgItemText(hDlg, IDC_FPS, szBuffer);
			SendDlgItemMessage(hDlg, IDC_PROGRESS, PBM_SETPOS, (int)(Bitrate_Monitor*Frame_Rate/(playback-Old_Playback))>>7, 0);

			Bitrate_Monitor = 0;
			process.mi = process.ed;
			Old_Playback = playback;
		}
	}

	if (AVI_Flag && avi_size>=MAX_AVI_SIZE)
		AVIKill();
}

static void Store_RGB24(unsigned char *src[], DWORD frame)
{
	if (AVI_Flag && AVI_Init)
	{
		if (!frame)
		{
			compvars.cbSize = sizeof(compvars);

			if (!ICCompressorChoose(hWnd, ICMF_CHOOSE_DATARATE | ICMF_CHOOSE_KEYFRAME,
				&birgb, NULL, &compvars, NULL))
				ThreadKill();

			// set AVI header
			ZeroMemory(&strinfo, sizeof(AVISTREAMINFO));
			strinfo.fccType					= streamtypeVIDEO;
			strinfo.fccHandler				= compvars.fccHandler;
			strinfo.dwQuality				= -1;
			strinfo.dwSuggestedBufferSize	= birgb.biSizeImage;
			strinfo.dwScale	= 1000;
			strinfo.dwRate = (unsigned int)((FO_Flag==FO_FILM) ? frame_rate*800 : frame_rate*1000);

			// set AVI save options
			opts.cbFormat = sizeof(birgb);
			opts.fccType = streamtypeVIDEO;
			opts.fccHandler = compvars.fccHandler;
			opts.dwKeyFrameEvery = compvars.lKey;
			opts.dwQuality = compvars.lQ;
			opts.dwBytesPerSecond = compvars.lDataRate<<10;
			opts.dwFlags = AVICOMPRESSF_VALID | AVICOMPRESSF_KEYFRAMES | AVICOMPRESSF_DATARATE;
			opts.dwInterleaveEvery = 0;
			opts.lpFormat = &birgb;
			opts.cbParms = compvars.cbState;
			opts.lpParms = compvars.lpState;
		}

		AVI_Init = false; avi_size = 0;
		AVIFileInit();

		sprintf(VideoOut, "%s.%02d.avi", szOutput, avi_count++);

		if (AVIFileOpen(&pfile, VideoOut, OF_WRITE | OF_CREATE, NULL) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (AVIFileCreateStream(pfile, &ps, &strinfo) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (AVIMakeCompressedStream(&psCompressed, ps, &opts, NULL) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (AVIStreamSetFormat(psCompressed, 0, &birgb, birgb.biSize) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}
	}

	if (chroma_format==CHROMA420)
	{
		conv420to422(src[1], u422, frame_type);
		conv420to422(src[2], v422, frame_type);

		conv422to444(u422, u444);
		conv422to444(v422, v444);
	}
	else
	{
		conv422to444(src[1], u444);
		conv422to444(src[2], v444);
	}

	if (Luminance_Flag)
	{
		LuminanceFilter(src[0], lum);
		y444 = lum;
	}
	else
		y444 = src[0];

	if (BFB)
	{
		conv444toRGB24odd(y444, u444, v444, rgb24);
		TFB = true;
		FlushRGB24();

		conv444toRGB24even(y444, u444, v444, rgb24);
		BFB = true;
		FlushRGB24();
	}
	else
	{
		conv444toRGB24even(y444, u444, v444, rgb24);
		BFB = true;
		FlushRGB24();

		conv444toRGB24odd(y444, u444, v444, rgb24);
		TFB = true;
		FlushRGB24();
	}

	if (FO_Flag!=FO_FILM && RFF)
		if (TFF)
		{
			TFB = true;
			FlushRGB24();
		}
		else
		{
			BFB = true;
			FlushRGB24();
		}
}


static void FlushRGB24()
{
	if (TFB & BFB)
	{
		if (AVI_Flag)
		{
			if (AVIStreamWrite(psCompressed, frame_count++, 1, rgb24,
				birgb.biSizeImage, 0, NULL, &frame_size) != AVIERR_OK)
			{
				AVIKill();
				ThreadKill();
			}

			avi_size += frame_size;
		}

		ShowFrame(false);

		playback++;
		TFB = BFB = false;
	}
}

static void Store_YUY2(unsigned char *src[], DWORD frame)
{
	if (AVI_Flag && AVI_Init)
	{
		if (!frame)
		{
			compvars.cbSize = sizeof(compvars);

			if (!ICCompressorChoose(hWnd, ICMF_CHOOSE_DATARATE | ICMF_CHOOSE_KEYFRAME,
				&biyuv, NULL, &compvars, NULL))
				ThreadKill();

			// set AVI header
			ZeroMemory(&strinfo, sizeof(AVISTREAMINFO));
			strinfo.fccType					= streamtypeVIDEO;
			strinfo.fccHandler				= compvars.fccHandler;
			strinfo.dwQuality				= -1;
			strinfo.dwSuggestedBufferSize	= biyuv.biSizeImage;
			strinfo.dwScale	= 1000;
			strinfo.dwRate = (unsigned int)((FO_Flag==FO_FILM) ? frame_rate*800 : frame_rate*1000);

			// set AVI save options
			opts.cbFormat = sizeof(biyuv);
			opts.fccType = streamtypeVIDEO;
			opts.fccHandler = compvars.fccHandler;
			opts.dwKeyFrameEvery = compvars.lKey;
			opts.dwQuality = compvars.lQ;
			opts.dwBytesPerSecond = compvars.lDataRate<<10;
			opts.dwFlags = AVICOMPRESSF_VALID | AVICOMPRESSF_KEYFRAMES | AVICOMPRESSF_DATARATE;
			opts.dwInterleaveEvery = 0;
			opts.lpFormat = &biyuv;
			opts.cbParms = compvars.cbState;
			opts.lpParms = compvars.lpState;

			iccf.dwRate = strinfo.dwRate;
			iccf.dwScale = strinfo.dwScale;
			iccf.lQuality = compvars.lQ;
			iccf.lDataRate = compvars.lDataRate<<10;
			iccf.lKeyRate = compvars.lKey;

			ICSendMessage(compvars.hic, ICM_COMPRESS_FRAMES_INFO, (WPARAM)&iccf, (DWORD)sizeof(ICCOMPRESSFRAMES));
		}

		AVI_Init = false; avi_size = 0;
		AVIFileInit();

		sprintf(VideoOut, "%s.%02d.avi", szOutput, avi_count++);

		if (AVIFileOpen(&pfile, VideoOut, OF_WRITE | OF_CREATE, NULL) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (AVIFileCreateStream(pfile, &ps, &strinfo) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (AVIMakeCompressedStream(&psCompressed, ps, &opts, NULL) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (AVIStreamSetFormat(psCompressed, 0, &birgb, birgb.biSize) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (!frame && !ICSeqCompressFrameStart(&compvars, (LPBITMAPINFO)&biyuv))
		{
			AVIKill();
			ThreadKill();
		}
	}

	if (chroma_format==CHROMA420)
	{
		conv420to422(src[1], u422, frame_type);
		conv420to422(src[2], v422, frame_type);
	}
	else
	{
		u422 = src[1];
		v422 = src[2];
	}

	if (Luminance_Flag)
	{
		LuminanceFilter(src[0], lum);
		y444 = lum;
	}
	else
		y444 = src[0];

	if (BFB)
	{
		conv422toyuy2odd(y444, u422, v422, yuy2);
		TFB = true;
		FlushYUY2();

		conv422toyuy2even(y444, u422, v422, yuy2);
		BFB = true;
		FlushYUY2();
	}
	else
	{
		conv422toyuy2even(y444, u422, v422, yuy2);
		BFB = true;
		FlushYUY2();

		conv422toyuy2odd(y444, u422, v422, yuy2);
		TFB = true;
		FlushYUY2();
	}

	if (FO_Flag!=FO_FILM && RFF)
		if (TFF)
		{
			TFB = true;
			FlushYUY2();
		}
		else
		{
			BFB = true;
			FlushYUY2();
		}
}

static void FlushYUY2()
{
	void *yuy2c;
	int key_flag;
	long yuy2c_size;

	if (TFB & BFB)
	{
		if (AVI_Flag)
		{
			if ((yuy2c = ICSeqCompressFrame(&compvars, 0, yuy2, &key_flag, &yuy2c_size)) == NULL)
			{
				AVIKill();
				ThreadKill();
			}

			if (AVIStreamWrite(ps, frame_count++, 1, yuy2c, yuy2c_size,
				key_flag ? AVIIF_KEYFRAME : 0, NULL, &frame_size) != AVIERR_OK)
			{
				AVIKill();
				ThreadKill();
			}

			avi_size += frame_size;
		}

		ShowFrame(false);

		playback++;
		TFB = BFB = false;
	}
}

int DetectVideoType(int frame, int trf)
{
	static int Old_TRF, Repeat_On, Repeat_Off;
	static bool Repeat_Init;

	if (frame)
	{
		if ((trf & 3) == ((Old_TRF+1) & 3))
			FILM_Purity++;
		else
			NTSC_Purity++;
	}
	else
	{
		FILM_Purity = NTSC_Purity = Repeat_On = Repeat_Off = 0;
		Repeat_Init = false;
	}

	Old_TRF = trf;

	if (trf & 1)
		Repeat_On ++;
	else
		Repeat_Off ++;

	if (Repeat_Init)
	{
		if (Repeat_Off-Repeat_On == 5)
		{
			Repeat_Off = Repeat_On = 0;
			return 0;
		}
		else if (Repeat_On-Repeat_Off == 5)
		{
			Repeat_Off = Repeat_On = 0;
			return 2;
		}
	}
	else
	{
		if (Repeat_Off-Repeat_On == 3)
		{
			Repeat_Off = Repeat_On = 0;
			Repeat_Init = true;
			return 0;
		}
		else if (Repeat_On-Repeat_Off == 3)
		{
			Repeat_Off = Repeat_On = 0;
			Repeat_Init = true;
			return 2;
		}
	}

	return 1;
}

static void AVIKill()
{
	AVI_Init = true; frame_count = 0;

	if (ps)
		AVIStreamClose(ps);

	if (psCompressed)
		AVIStreamClose(psCompressed);

	if (pfile)
		AVIFileClose(pfile);

	if (Stop_Flag || Fault_Flag)
	{
		AVIFileExit();

		if (Store_Flag!=STORE_RGB24)
			ICSeqCompressFrameEnd(&compvars);

		ICCompressorFree(&compvars);
	}
}

// These guys defined in getbit.c
extern char NextNum_Buttons;
extern char Num_Buttons;
extern unsigned short StartY[36]; 
extern unsigned short EndY[36];	
extern unsigned short StartX[36];
extern unsigned short EndX[36];
extern int DrawButtons ;
extern int ButtonsChanged;	

void DisplayButtons()
{
	// This function displays the button rectangles as defined in the navpacks...
	HPEN hpen;
	char i;
	if(!DrawButtons)
		return;
//OutputDebugString("DISPLAYBUTTONS \n"); // Use this to see when this function is called.
	if(ButtonsChanged && Store_Flag==STORE_YUY2)
	{
		// This causes the entire window rectangle to be redrawn, which means
		// the previous rectangles will be erased and replaced by the new.
		// This is only needed in overlay mode because the bitmap mode erases
		// the rectangles anyway.
		RECT Complete;
		Complete.left = Complete.top = 0; 
		Complete.right = (int)(Coded_Picture_Width * YUVScaleX);
		Complete.bottom = (int)(Vert_Display_Size * YUVScaleY);;
		InvalidateRect(hWnd,&Complete,true);
	}
	hpen = CreatePen(PS_SOLID, 1, ButtonColor);
	SelectObject(hDC, hpen);
	SelectObject(hDC, GetStockObject(HOLLOW_BRUSH)); 
	for(i=0;i<Num_Buttons;i++)
	{
		Rectangle(hDC,(long) (StartX[i] * YUVScaleX),(long) (StartY[i] * YUVScaleY),
			(long) (EndX[i] * YUVScaleX),(long) (EndY[i] * YUVScaleY));
	}
	ButtonsChanged=0;
}

static bool DisplayWasErased = false; // I need this to remember to invalidate the entire rect.
// This can be used to erase the video display.
void EraseFrame(COLORREF Color)
{
	RECT prect;

	if(AutoScale == 1)
	{
		// If AutoScale is 1, we scale the y-axis according to the aspect ratio, and according
		// to the x-axis scaling.
		int Dar = (ForcedDAR == -1) ? aspect_ratio_information : ForcedDAR;
		switch(Dar)
		{
		case 0:
		case 1: YUVScaleY = YUVScaleX * (float)Coded_Picture_Width/Vert_Display_Size; break;
		case 2: YUVScaleY = YUVScaleX / 1.33333f * (float)Coded_Picture_Width/Vert_Display_Size; break;
		case 3: YUVScaleY = YUVScaleX / 1.77778f * (float)Coded_Picture_Width/Vert_Display_Size; break;
		case 4: YUVScaleY = YUVScaleX / 2.21f * (float)Coded_Picture_Width/Vert_Display_Size; break;
		}
	}

	SetRect(&prect,0,0,(int)(Coded_Picture_Width * YUVScaleX),(int)(Vert_Display_Size * YUVScaleY));
	FillRect(hDC,&prect,(HBRUSH)CreateSolidBrush(Color));
	// This will be required in overlay mode where the rectangle needs to be invalidate for the video
	// to be displayed.
	DisplayWasErased = true;
}

extern void MixFunction(unsigned char* Bitmap, long Width, long Height, void* Param);

int WSDisplayMode = 0;

void ShowFrame(bool move)
{
	HRESULT hr;
	if (!Check_Flag)
		return;

// erase background
//SelectObject(hDC, CreateSolidBrush(RGB(0, 0, 0)));
//PatBlt(hDC, 0, 0, 400, 400, PATCOPY);

	if(AutoScale == 1)
	{
		// If AutoScale is 1, we scale the y-axis according to the aspect ratio, and according
		// to the x-axis scaling.
		int Dar = (ForcedDAR == -1) ? aspect_ratio_information : ForcedDAR;
		switch(Dar)
		{
		case 0:
		case 1: YUVScaleY = YUVScaleX * (float)Coded_Picture_Width/Vert_Display_Size; break;
		case 2: YUVScaleY = YUVScaleX / 1.33333f * (float)Coded_Picture_Width/Vert_Display_Size; break;
		case 3: YUVScaleY = YUVScaleX / 1.77778f * (float)Coded_Picture_Width/Vert_Display_Size; break;
		case 4: YUVScaleY = YUVScaleX / 2.21f * (float)Coded_Picture_Width/Vert_Display_Size; break;
		}
	}

	if (Store_Flag==STORE_YUY2)
	{
		if ((Display_Flag == 1 || move) && DDOverlay_Flag)
		{
			POINT point = {0, 0};
			RECT orect, prect;

			if (!move && (NeverUseOverlay || IDirectDrawSurface_Lock(lpOverlay, NULL, &ddsd, 0, NULL)==DD_OK))
			{
				int i;
				unsigned char *src = yuy2;		
				unsigned char *dst = (unsigned char *)ddsd.lpSurface;

				for (i=0; i<Clip_Height; i++)
				{
					memcpy(dst, src, DOUBLE_CLIP_WIDTH);
					src += DOUBLE_CLIP_WIDTH;
					dst += ddsd.lPitch;
				}
				if(!NeverUseOverlay)
					IDirectDrawSurface_Unlock(lpOverlay, NULL);
			}
			SetRect(&orect, 0, 0, Coded_Picture_Width, Vert_Display_Size);
			ClientToScreen(hWnd, &point);
			prect.left = point.x;
			prect.right = point.x + (int)(Coded_Picture_Width * YUVScaleX);
			prect.top = point.y;
			prect.bottom = point.y + (int)(Vert_Display_Size * YUVScaleY);

			if (prect.left < 0)
			{
				orect.left = -prect.left;
				prect.left = 0;
			}

			if (prect.top < 0)
			{
				orect.top = -prect.top;
				prect.top = 0;
			}

			if (prect.right > GetSystemMetrics(SM_CXSCREEN))
			{
				orect.right = Coded_Picture_Width + GetSystemMetrics(SM_CXSCREEN) - prect.right;
				prect.right = GetSystemMetrics(SM_CXSCREEN);
			}

			if (prect.bottom > GetSystemMetrics(SM_CYSCREEN))
			{
				orect.bottom = Vert_Display_Size + GetSystemMetrics(SM_CYSCREEN) - prect.bottom;
				prect.bottom = GetSystemMetrics(SM_CYSCREEN);
			}

			// The hardware does the rescaling here!
			DisplayButtons();
			if(DisplayWasErased)
			{
				// I'm invalidating the entire area to refresh the display. 
				InvalidateRect(hWnd,0,true);
				DisplayWasErased = false;
			}
			if(!NeverUseOverlay)
				hr = IDirectDrawSurface_UpdateOverlay(lpOverlay, &orect, lpPrimary, &prect,
					DDOVER_SHOW | DDOVER_DDFX | DDOVER_KEYDESTOVERRIDE, &ddofx);
//			IDirectDrawSurface_SetOverlayPosition(lpOverlay, prect.left, prect.top); // Original comment.
		}
	}
	else
	{
		unsigned char *source = rgb24;
		if (Display_Flag == 1 || move)
		{
			unsigned char *ScaledDisplay = rgb24Scaled;
			static int ScaledWidth = 0;
			static int ScaledHeight = 0;
			static float LastYUVScaleX = 0;
			static float LastYUVScaleY = 0;
			// Let's see if we do need to rescale this picture. If we're only refreshing, NeedsScaling will
			// be 0, but the X/Y scaling ratios can still have been modified.
			NeedsScaling = NeedsScaling | (LastYUVScaleX != YUVScaleX );
			NeedsScaling = NeedsScaling | (LastYUVScaleY != YUVScaleY );
			LastYUVScaleX = YUVScaleX;
			LastYUVScaleY = YUVScaleY;

			if((YUVScaleX == 1) && (YUVScaleY == 1))
			{
				// If we're not rescaling we can just point to the original array.
				ScaledWidth = Clip_Width, ScaledHeight = Clip_Height;
				ScaledDisplay = rgb24;
			}

			// I'm moving this guy here, so resizing can happen even when we're only refreshing an already
			// decoded picture. 
			ResizeWindow(Clip_Width, Vert_Display_Size);
			if(BitmapMixFunction)
			{
				if(rgb24Mixed == 0)
					rgb24Mixed = (unsigned char*)_aligned_malloc((unsigned int)(Coded_Picture_Width*Coded_Picture_Height*3), 64);

				// Call the bitmap mix function.
				memcpy(rgb24Mixed,rgb24,3*Coded_Picture_Width*Clip_Height);
				BitmapMixFunction(rgb24Mixed, Coded_Picture_Width, Clip_Height, BitmapMixParam,0);
				NeedsScaling = 1;
				source = rgb24Mixed;
			}
			if(NeedsScaling && (YUVScaleX != 1 || YUVScaleY != 1))
			{
				// Rescaling code. Ideally, I should calculate horizontal and vertical scaling
				// mapping arrays once and for all, not for each frame! But that's OK. 
				// NO! Ideally, I should use StretchDIBits() which I didn't know existed!
				unsigned char *dest = ScaledDisplay;
				unsigned char *src = source;
				float FactX;
				float FactY;
				int i,j;
				int StartX = 0;

				// Learned the hard way! The width of a bmp picture HAS to be a multiple of 4!!!
				FactX = 1/YUVScaleX;
				FactY = 1/YUVScaleY;
				// In letterbox, we leave 1/8 of the height above and below in black. We need an
				// additional vertical shrinking of 3/4
				if(WSDisplayMode == 1)
					FactY *= 4.0f/3;
				// We have to modify FactX so the resulting width will be a multiple of 4! BMP imposes that!!!
				ScaledWidth = (int)(Coded_Picture_Width/FactX);
				// Then round that guy to the nearest multiple of 4.
				ScaledWidth = 4*((int)((float)ScaledWidth/4+0.5));
				// Look at the loop below, xindex = (int)(j * FactX + .5f); When i = NewWidth-1, this should
				// point to the last pixel, xindex = Clip_Width-1. Recalculate FactX to ensure that.
				FactX = ((float) Coded_Picture_Width - 1 - 0.5f) / (ScaledWidth-1);
				ScaledHeight = (int)(Vert_Display_Size * YUVScaleY);
				if(WSDisplayMode == 1)
				{
					// In Letterbox, fill the 1/8th top with black pixels;
					memset(dest,0,3*ScaledWidth*((int)(ScaledHeight/8)));
					dest += 3*ScaledWidth*((int)(ScaledHeight/8));
				}
				if(WSDisplayMode == 2)
				{
					// In Pan&Scan, we stretch the picture horizontally by 4/3, and we dump 1/4th of
					// the original bitmap (1/8 on the left, 1/8 on the right).
					FactX /= 4.0f/3; 
					StartX = Coded_Picture_Width / 8;
				}
				// Loop over lines:
				for(i=0;;i++)
				{
					// Find start of new line. bmp files are stored bottom to top apparently! ARGH! Since 
					// Vert_Display_Size can be smaller than Coded_Picture_Height, we need to skip the bottom.
					int yindex = Clip_Height-Vert_Display_Size + (int)(i * FactY + .5f);
					if(yindex >= Coded_Picture_Height)
						break;
					// The factor 3 comes from 3 bytes (R,B,G).
					src = source + yindex * Coded_Picture_Width * 3;
					// Copy this line.
					for(j=0;j<ScaledWidth;j++)
					{
						// Find pixel location in line.
						int xindex = (int)(j * FactX + .5f + StartX);
						if(xindex >= Coded_Picture_Width)
							break;
						// Copy the three BGR components for that pixel.
						*dest++ = src[3*xindex];
						*dest++ = src[3*xindex+1];
						*dest++ = src[3*xindex+2];
					}
				}
				// And fill the 1/8th bottom with black pixels;
				if(WSDisplayMode == 1)
				{
					memset(dest,0,3*ScaledWidth*((int)(ScaledHeight/8)));
				}
				NeedsScaling = 0;
				birgb.biWidth = ScaledWidth;
				birgb.biHeight = ScaledHeight;
				birgb.biSizeImage = 3*ScaledWidth*ScaledHeight;
			}
			// Call this after rescaling, pass 1 as last argument to indicate post-scaling.
			if(BitmapMixFunction)
				BitmapMixFunction(ScaledDisplay, ScaledWidth, ScaledHeight, BitmapMixParam,1);


#if 0
			// This kind of works, but creates a crappy image (don't know why) when rescaling. Also,
			// there isn't any computation savings that I can see... Not sure why either...
			birgb.biWidth = Coded_Picture_Width;
			birgb.biHeight = Vert_Display_Size;
			birgb.biSizeImage = 3*Coded_Picture_Width*Vert_Display_Size;
			StretchDIBits(
			  hDC,                // handle to device context
			  0,              // x-coordinate of upper-left corner of dest. rectangle
			  0,              // y-coordinate of upper-left corner of dest. rectangle
//			  (int)(Coded_Picture_Width * YUVScaleX),         // width of destination rectangle
//			  (int)(Vert_Display_Size * YUVScaleY),        // height of destination rectangle
			  (int)(ScaledWidth),         // width of destination rectangle
			  (int)(ScaledHeight),        // height of destination rectangle
			  0,               // x-coordinate of upper-left corner of source rectangle
			  0,               // y-coordinate of upper-left corner of source rectangle
			  Coded_Picture_Width,          // width of source rectangle
			  Vert_Display_Size,         // height of source rectangle
			  rgb24,            // address of bitmap bits
			  (LPBITMAPINFO)&birgb,  // address of bitmap data
			  DIB_RGB_COLORS,                   // usage flags
			  SRCCOPY                    // raster operation code
			);
#else
// I could use this to do the pan&scan display.
//			IntersectClipRect(hDC,0,0,100,100);
			SetDIBitsToDevice(hDC, 0, 0, ScaledWidth, ScaledHeight,
				0, 0, 0, ScaledHeight, ScaledDisplay, (LPBITMAPINFO)&birgb, DIB_RGB_COLORS);
#endif

			DisplayButtons();
			// Call the PostDisplay function is the pointer isn't null.
			if(PostDisplay)
				PostDisplay(hDC,PostDisplayParam);
		}
	}
}
